
#ifndef PLS_TBB_LIKE_TASK_H
#define PLS_TBB_LIKE_TASK_H

#include <pls/internal/base/aligned_stack.h>
#include "abstract_task.h"
#include "thread_state.h"

namespace pls {
    namespace internal {
        namespace scheduling {
            class tbb_task;
            class tbb_sub_task {
                friend class tbb_task;

                // Coordinate finishing of sub_tasks
                std::atomic_uint32_t ref_count_;
                tbb_sub_task* parent_;

                // Access to TBB scheduling environment
                tbb_task* tbb_task_;

                // Double-Ended Queue management
                tbb_sub_task* below_;
                tbb_sub_task* above_;

                // Stack Management (reset stack pointer after wait_for_all() calls)
                base::aligned_stack::state stack_state_;
            protected:
                explicit tbb_sub_task();
                tbb_sub_task(const tbb_sub_task& other);

                virtual void execute_internal() = 0;
                // SubClass Implementations:
                // Do Work
                // |-- Spawn Sub Task (new subtask; spawn(subtask);)
                // |-- Spawn Sub task
                // Do Work
                // |-- Wait For All
                // Do  Work
                // |-- Spawn Sub Task

                template<typename T>
                void spawn_child(const T& sub_task);
                void wait_for_all();
            private:
                void spawn_child_internal(tbb_sub_task* sub_task);
                void execute();

            public:
                virtual void test() {
                    std::cout << "Test" << std::endl;
                }
            };

            class tbb_task: public abstract_task {
                friend class tbb_sub_task;

                tbb_sub_task* root_task_;
                base::aligned_stack* my_stack_;

                // Double-Ended Queue management
                tbb_sub_task* top_;
                tbb_sub_task* bottom_;

                // Steal Management
                tbb_sub_task* last_stolen_;

                tbb_sub_task* get_local_sub_task();
                tbb_sub_task* get_stolen_sub_task();

                bool internal_stealing(abstract_task* other_task) override;
                bool split_task() override;

            public:
                explicit tbb_task(tbb_sub_task* root_task):
                        abstract_task{0, 0},
                        root_task_{root_task},
                        top_{nullptr},
                        bottom_{nullptr},
                        last_stolen_{nullptr} {
                    my_stack_ = base::this_thread::state<thread_state>()->task_stack_;
                    root_task_->tbb_task_ = this;
                    root_task_->stack_state_ = my_stack_->save_state();
                };

                void execute() override {
                    root_task_->execute();
                }
            };

            template<typename T>
            void tbb_sub_task::spawn_child(const T& task)  {
                static_assert(std::is_base_of<tbb_sub_task, T>::value, "Only pass tbb_sub_task subclasses!");

                T* new_task = tbb_task_->my_stack_->push(task);
                spawn_child_internal(new_task);
            }
        }
    }
}

#endif //PLS_TBB_LIKE_TASK_H
